#  CampusIO.pm - a FlowScan-derived class for reporting on campus traffic I/O
#  Copyright (C) 1998-2000  Dave Plonka
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

# $Id: CampusIO.pm,v 1.45 2000/10/18 21:10:57 dplonka Exp $
# Dave Plonka <plonka@doit.wisc.edu>

use strict;

package CampusIO;

require 5;
require Exporter;
require FlowScan;

@CampusIO::ISA=qw(FlowScan Exporter);
# convert the RCS revision to a reasonable Exporter VERSION:
'$Revision: 1.45 $' =~ m/(\d+)\.(\d+)/ && (( $CampusIO::VERSION ) = sprintf("%d.%03d", $1, $2));

=head1 NAME

CampusIO -

=head1 SYNOPSIS

   $ flowscan CampusIO

=head1 DESCRIPTION

The following methods and subroutines are defined:

=over 4

=cut

use FindBin;
use Cflow qw(:flowvars 1.024); # for use in wanted sub
use RRDs;
use Boulder::Stream;
use Socket; # for inet_ntoa, inet_aton
use POSIX; # for UINT_MAX, mktime
use IO::File;
use File::Basename;
use ConfigReader::DirectiveStyle;
use Net::Patricia 1.008;

my $c = new ConfigReader::DirectiveStyle;
$c->directive('NextHops');
$c->directive('OutputIfIndexes');
$c->required('OutputDir');
$c->directive('TCPServices');
$c->directive('UDPServices');
$c->directive('Protocols');
$c->directive('ASPairs');
$c->directive('LocalNextHops');
$c->directive('LocalSubnetFiles');
$c->directive('Rateup');
$c->directive('Verbose');
$c->directive('NapsterSubnetFiles');
$c->directive('NapsterSeconds');
$c->directive('NapsterPorts');
$c->load("${FindBin::Bin}/CampusIO.cf");

if (1 <= $c->value('Verbose')) {
   $CampusIO::verbose = 1
}
if (2 <= $c->value('Verbose')) {
   $CampusIO::Verbose = 1
}

# outputdir can be absolute or relative (to the flow file's directory):
$CampusIO::outputdir = $c->value('OutputDir');
$CampusIO::rateup = $c->value('Rateup');

$CampusIO::thingy; # this is a global set by report subroutine and used by the sort subs
$CampusIO::thing = 'bytes';

# { these vars are used by the wanted subroutine:

foreach my $if (split(m/\s*,\s*/, $c->value('OutputIfIndexes'))) {
   push(@CampusIO::output_if, $if)
}

$CampusIO::hops_ptrie = new Net::Patricia;
die unless ref($CampusIO::hops_ptrie);

my $hop;
foreach $hop (split(m/\s*,\s*/, $c->value('NextHops'))) {
   push(@CampusIO::hops, unpack("N", inet_aton($hop)));
   if (!$CampusIO::hops_ptrie->add_string($hop)) {
      warn "hops_ptrie->add of $hop failed!\n"
   }
}

if (!@CampusIO::output_if && !@CampusIO::hops) {
   die "No valid OutputIfIndexes or NextHops (One or the other is REQUIRED)!"
}

$CampusIO::localhops_ptrie = new Net::Patricia;
die unless ref($CampusIO::localhops_ptrie);

foreach $hop (split(m/\s*,\s*/, $c->value('LocalNextHops'))) {
   if (!$CampusIO::localhops_ptrie->add_string($hop)) {
      warn "localhops_ptrie->add of $hop failed!\n"
   }
}

# { Handle interesting services...
my %services = (tcp => 'TCPServices',
		udp => 'UDPServices');
while (my($protoname, $option) = each(%services)) {
   my $service;
   my $port;
   my $proto = scalar(getprotobyname($protoname));
   die("undefined protocol \"$protoname\"!") unless $proto;
   foreach $service (split(m/\s*,\s*/, $c->value($option))) {
      if ($service !~ m/^\d+$/) {
         $port = getservbyname($service, $protoname);
         $port || die "undefined $protoname service \"$service\"!"
      } else {
	 $port = $service
      }
      $CampusIO::service{$proto}{$port} = 1
   }
}
# }

{ # Handle interesting protocols...
   my $proto; # don't collide with imported $Cflow::protocol!
   foreach $proto (split(m/\s*,\s*/, $c->value('Protocols'))) {
      my $proto = scalar(getprotobyname($proto));
      die "undefined protocol \"$proto\"!" if (!$proto);
      $CampusIO::proto{$proto} = 1
   }
}

# These are the ASes for which we'll create ".rrd" files:
@CampusIO::as2as = split(m/\s*,\s*/, $c->value('ASPairs'));

# %CampusIO::RealServer will be a cache of hosts that we think are "Real"
# Servers based on having seen a flow involving their "well known" TCP ports.
# If we subsequently see traffic from one of these hosts involving the
# "well known" range of UDP ports we'll count that as "Real" Traffic.
%CampusIO::RealServer = ();

# Multicast stuff:
$CampusIO::MCAST_NET = unpack('N', inet_aton('224.0.0.0'));
$CampusIO::MCAST_MASK = unpack('N', inet_aton('240.0.0.0'));

# Handle Napster stuff if it's configured:
@CampusIO::napster_files = split(m/\s*,\s*/, $c->value('NapsterSubnetFiles'));
if (@CampusIO::napster_files &&
    (@CampusIO::napster_files = <@CampusIO::napster_files>)) {

   if (!(@CampusIO::NapsterPorts =
	    split(m/\s*,\s*/, $c->value('NapsterPorts')))) {
     # these are the defaults as of this writing (Mar  9 2000):
     @CampusIO::NapsterPorts = (4444, 5555, 6666, 6699, 7777, 8875, 8888);
     warn("NapsterPorts is unset... using defaults: ",
          join(", ", @CampusIO::NapsterPorts), "\n");
   }
   $CampusIO::NapsterSeconds = $c->value('NapsterSeconds');
   if (0 >= $CampusIO::NapsterSeconds) { # default to 1/2 hour if unset:
      $CampusIO::NapsterSeconds = 30*60; # minutes*seconds
      warn("NapsterSeconds is unset... using default: ",
	   $CampusIO::NapsterSeconds, "\n")
   }

   # { initialize the Napster Patricia Trie:
   $CampusIO::nptrie = new Net::Patricia;
   die unless ref($CampusIO::nptrie);

   @CampusIO::napster_files = <@CampusIO::napster_files>;
   my($file, $stream, $cargo);
   foreach $file (@CampusIO::napster_files) {
      print(STDERR "Loading \"$file\" ...\n") if -t;
      my $fh = new IO::File "<$file";
      $fh || die "open \"$file\", \"r\": $!\n";
      $stream = new Boulder::Stream $fh;
      while ($cargo = $stream->read_record) {
         my $subnet = $cargo->get('SUBNET');
         die unless $subnet;
	 my $collision;
	 if ($collision = $CampusIO::nptrie->match_string($subnet)) {
            warn "$subnet nptrie->add skipped - collided with $collision->{SUBNET}\n";
	    next
	 }
	 if (!$CampusIO::nptrie->add_string($subnet)) {
            warn "$subnet nptrie->add failed!\n";
	 }
      }
      undef $fh
   }
   # }
}

@CampusIO::subnet_files = split(m/\s*,\s*/, $c->value('LocalSubnetFiles'));

# { initialize the "local" Patricia Tree
# flows with a source that is not within the $ptrie will be considered to
# be candidate inbound flows...

@CampusIO::subnets_files = <@CampusIO::subnet_files>;
$CampusIO::ptrie = new Net::Patricia;
die unless ref($CampusIO::ptrie);
my($subnets_file, $stream, $cargo);
foreach $subnets_file (@CampusIO::subnets_files) {
   print(STDERR "Loading \"$subnets_file\" ...\n") if -t;
   my $fh = new IO::File "<$subnets_file";
   $fh || die "open \"$subnets_file\", \"r\": $!\n";
   $stream = new Boulder::Stream $fh;
   while ($cargo = $stream->read_record) {
      my $subnet = $cargo->get('SUBNET');
      my $hr = { SUBNET => $subnet };
      my $collision;
      if ($collision = $CampusIO::ptrie->match_string($subnet)) {
         warn "$subnet skipped.  It collided with $collision->{SUBNET}\n";
	 next
      }
      if ($CampusIO::ptrie->add_string($subnet, $hr)) {
         push(@CampusIO::subnets, $hr);
      } else {
         warn "$subnet add failed!\n";
	 next
      }
   }
   undef $fh
}

# }

# }

die("No subnets defined in subnet files? (\"", join('", "', @CampusIO::subnet_files), "\")\n") unless @CampusIO::subnets;

=item new

The B<new> method constructs and returns a B<CampusIO> object.

=cut

sub new {
   my $self = {};
   my $class = shift;
   return bless _init($self), $class
}

sub _init {
   my $self = shift;
   $self->{CampusIO}{wanted} = 0; # boolean value returned by wanted subroutine
   $self->{CampusIO}{which} = ''; # 'in'/'out' (valid when {wanted} is non-zero)

   # initialize the totals for "interesting" services ($self->{total}{service}):
   foreach my $protocol (keys %CampusIO::service) {
      foreach my $port (keys %{$CampusIO::service{$protocol}}) {
         foreach my $direction ('src', 'dst') {
            foreach my $which ('in', 'out') {
               $self->{total}{service}{$protocol}{$direction}{$port}{$which}{bytes} = 0;
               $self->{total}{service}{$protocol}{$direction}{$port}{$which}{pkts} = 0;
               $self->{total}{service}{$protocol}{$direction}{$port}{$which}{flows} = 0
            }
         }
      }
   }

   return $self
}

=item wanted

=cut

# Quake - apparently Quake 3 uses UDP with source and destination port of 27960

sub QuakeWanted {
   my $self = shift;
   my $which = shift;
   my $ref = shift;
   if (27960 == $srcport && 27960 == $dstport && 17 == $protocol) {
      $ref->{app}{Quake}{$which}{flows}++;
      $ref->{app}{Quake}{$which}{bytes} += $bytes;
      $ref->{app}{Quake}{$which}{pkts} += $pkts;
      return 1
   }
   return 0
}

# As of this writing, apparently Real Network's "RealMeadia" (Audio and/or
# Video) clients and servers uses these ports:
#
#    TCP port 7070 for connecting to pre-G2 RealServers
#    TCP port 554 and 7070 for connecting to G2 RealServers 
#    UDP ports 6970 - 7170 (inclusive) for incoming traffic only 
#
# Apparently the content can also be sent in HTTP format... which we
# couldn't differentiate from other HTTP traffic. Ugh.
#
# This was gleaned from "http://service.real.com/firewall/adminfw.html".

sub RealWanted {
   my $self = shift;
   my $which = shift;
   my $ref = shift;
   if (6 == $protocol && (7070 == $srcport || 554 == $srcport)) {
      # build a cache of hosts that look like "Real" servers:
      # (FIXME - this cache should be purged of old entries periodically.)
      $CampusIO::RealServer{$srcaddr} = $endtime;
      return 1
   } elsif (17 == $protocol && $dstport >= 6970 && $dstport <= 7170 &&
	    defined($CampusIO::RealServer{$srcaddr})) {
      $ref->{app}{RealAudio}{$which}{flows}++;
      $ref->{app}{RealAudio}{$which}{bytes} += $bytes;
      $ref->{app}{RealAudio}{$which}{pkts} += $pkts;
      return 1
   }
   return 0
}

# PASV mode ftp data

sub ftpPASVWanted {
   my $self = shift;
   my $which = shift;
   my $ref = shift;

   # FIXME?  Handle ftp-data in some way?  But how?
   return(0) unless 6 == $Cflow::protocol && (21 == $Cflow::srcport ||
				              21 == $Cflow::dstport ||
				              (1024 <= $Cflow::srcport &&
				               1024 <= $Cflow::dstport));

   # Only flows representing ftp (control) or TCP traffic on unprivileged
   # ports should get to this point in the code.

   my($client, # client IP address (as host-ordered integer)
      $server, # server IP address (as host-ordered integer)
      $direction); # 'src' or 'dst'

   return(0) unless 6 == $Cflow::protocol; # must be tcp

   # Skip pathalogical flows since $srcaddr == $dstaddr will break
   # "$client:$server" stuff below:
   return(0) if ($Cflow::srcaddr == $Cflow::dstaddr);

   # FIXME? What happens when two host simultaneous have ESTABLISHED ftp
   # command (and PASV mode data) streams between each other, negotiated
   # in opposite directions? (Something bad I think, esp. regarding 'src'
   # and 'dst'.)  For the time being, we'll assume "never happens."

   if (1024 <= $Cflow::srcport && 1024 <= $Cflow::dstport) {
      # At the point we've got unreserverd TCP ports for src and dst...
      # This *could* be a ftp PASV data flow...

      # See if we get a hit on the FTPSession, cache...
      my $r = $CampusIO::FTPSession{"$Cflow::srcaddr:$Cflow::dstaddr"};
      # If not, skip this candidate flow (it still might be PASV ftp, but
      # we can't tell since we have yet to see an ftp control TCP stream
      # between these hosts):
      return(0) if !ref($r) || (-1 != $r->[1] && $r->[1] <= $Cflow::endtime);

      warn "ftp-PASV data flow: $srcip.$srcport -> $dstip.$dstport $protocol $pkts $bytes\n" if $CampusIO::Verbose;

      $ref->{app}{"ftpPASV_$r->[2]"}{$which}{flows}++;
      $ref->{app}{"ftpPASV_$r->[2]"}{$which}{bytes} += $Cflow::bytes;
      $ref->{app}{"ftpPASV_$r->[2]"}{$which}{pkts} += $Cflow::pkts;

      if ($self->{ftpPASVFH}) {
	 syswrite($self->{ftpPASVFH}, $Cflow::raw, length $Cflow::raw)
      }

      return 1

   } elsif (21 == $dstport) {

      $server = $Cflow::dstaddr;
      $client = $Cflow::srcaddr

   } elsif (21 == $Cflow::srcport) {

      $server = $Cflow::srcaddr;
      $client = $Cflow::dstaddr

   } else {
      return 0
   }

   # At this point we think we have an ftp control flow (using TCP port 21)...
   if ($Cflow::TH_ACK & $Cflow::tcp_flags) {
       # thanks to Simon Leinen <simon@limmat.switch.ch> for hint to look
       # for ACK in TCP stream, to be sure it is an active session:
      if (-1 == $CampusIO::FTPSession{"$client:$server"}[0] ||
	  $CampusIO::FTPSession{"$client:$server"}[0] < $Cflow::endtime) {
	 # We don't want to be too liberal with what we presume are PASV
	 # ftp data flows since users of other sharing applications (such
	 # as gnutella) may be ftp "power users" as well, and could have
	 # both running simultaneously, which would inadvertently lead us
	 # to believe that they are the result of PASV ftp transfers.

         # [0] stores the time_t value for the last flow seen  with ACK set
         # [1] stores the time_t value for the last flow seen with FIN set
         # [2] stores the "direction", either 'src' or 'dst' (meaning that
	 #     the ftp server is either the source or destination of traffic
	 #     for flows that get a hit using
	 #     $CampusIO::FTPSession{"$srcaddr:$dstaddr"}.
	 # (Traditionally (time_t)-1 is an invalid time_t value.)
	 $CampusIO::FTPSession{"$client:$server"}[0] = $Cflow::endtime;
	 $CampusIO::FTPSession{"$client:$server"}[1] = -1; # no FIN seen yet
	 $CampusIO::FTPSession{"$client:$server"}[2] = 'dst';
	 # Duplicate the entry in the hash (with the client and server
	 # address reversed in the key, so that we only have to do one
	 # hash lookup when testing candidate ftp PASV data flows:
	 $CampusIO::FTPSession{"$server:$client"}[0] = $Cflow::endtime;
	 $CampusIO::FTPSession{"$server:$client"}[1] = -1; # no FIN seen yet
	 $CampusIO::FTPSession{"$server:$client"}[2] = 'src';
      }
   }
   if ($Cflow::TH_FIN & $tcp_flags) {
      # lose the ACK time, we've got FIN:
      $CampusIO::FTPSession{"$client:$server"}[0] = -1;
      $CampusIO::FTPSession{"$client:$server"}[1] = $Cflow::endtime; # FIN time
      $CampusIO::FTPSession{"$client:$server"}[2] = 'dst';
      # Duplicate the entry in the hash (with the client and server
      # address reversed in the key), so that we only have to do one
      # hash lookup when testing candidate ftp PASV data flows:
      # lose the ACK time, we've got FIN:
      $CampusIO::FTPSession{"$server:$client"}[0] = -1;
      $CampusIO::FTPSession{"$server:$client"}[1] = $Cflow::endtime; # FIN time
      $CampusIO::FTPSession{"$server:$client"}[2] = 'src';
   }

   # FIXME? Should we be keeping a count of the number of ftp control streams
   # between a given client and server?  (While it is probably unlikely for
   # this to occur, the code certainly can't do the right thing unless it
   # keeps count of the current number of ftp control streams.)

   return 0
}

# Napster
#
#    TCP port 8875 to connect to napster.com servers:
#
#       208.49.228.0/255.255.255.0
#       208.184.216.0/255.255.255.0
#       208.49.239.240/255.255.255.240
#       208.178.175.128/255.255.255.248
#       208.178.163.56/255.255.255.248
#
#    any TCP port to connect to other napster users' servers,
#    unreserved ports are the most likely(?).  Defaults are
#    4444,5555,6666,7777,8888.  I have read that ports 6697 and
#    6699 are used as well.

sub NapsterWanted {
   my $self = shift;
   my $which = shift;
   my $ref = shift;
   my($outside_addr, $inside_addr);

   # Skip TCP traffic on reserved ports:
   # We check for the protocol here since some Napster traffic may be ICMP, no?
   # FIXME? should the following test for reserved ports be "1024 <= $val"?
   return(0) unless (1 == $protocol ||
		     (6 == $protocol && 1024 < $srcport && 1024 < $dstport));

   if ('in' eq $which) {
      $outside_addr = $srcaddr;
      $inside_addr = $dstaddr
   } elsif ('out' eq $which) {
      $outside_addr = $dstaddr;
      $inside_addr = $srcaddr
   } else {
      die
   }

   # FIXME - we should only consider it to be an "active" NapServer
   # (either a redirect or index server) if it is the *source* and
   # has the ACK bit set:  
   if (6 == $protocol &&
       ($Cflow::TH_ACK & $tcp_flags) &&
       $CampusIO::nptrie->match_integer($outside_addr)) {
      # OK, this looks like traffic involving a a NapServer...
      # build a cache of hosts that look like "Napster" servers:
      # Periodically, these caches are purged of old entries by the
      # NapsterPurgeCache subroutine.
      $CampusIO::NapServer{$outside_addr} = $endtime;
      $CampusIO::NapUser{$inside_addr} = $endtime;
   } elsif (defined($CampusIO::NapUser{$inside_addr})) {
      # FIXME - check $outside_port against the default ports (above) and
      # report "high" and "low" confidence traffic accordingly.
   } else {
      return 0
   }
   if (((grep($srcport == $_, @CampusIO::NapsterPorts) ||
         grep($dstport == $_, @CampusIO::NapsterPorts)) &&
        6 == $protocol) ||
	(1 == $protocol && 28 == $bytes/$pkts)) {
      # In addition the the TCP streams to pass data,
      # I've seen Napster application users doing lots of these 28-byte "pings".
      # "Confidence is high -- I repeat -- Confidence is high."
      warn "HIGH confidence NapFlow: $srcip.$srcport -> $dstip.$dstport $protocol $pkts $bytes\n" if $CampusIO::Verbose;
      $ref->{app}{NapUser}{$which}{flows}++;
      $ref->{app}{NapUser}{$which}{bytes} += $bytes;
      $ref->{app}{NapUser}{$which}{pkts} += $pkts;
      if ($self->{NapUserFH}) {
	 syswrite($self->{NapUserFH}, $Cflow::raw, length $Cflow::raw)
      }
   } else { # low/no confidence...
      warn " LOW confidence NapFlow: $srcip.$srcport -> $dstip.$dstport $protocol $pkts $bytes\n" if $CampusIO::Verbose;
      $ref->{app}{NapUserMaybe}{$which}{flows}++;
      $ref->{app}{NapUserMaybe}{$which}{bytes} += $bytes;
      $ref->{app}{NapUserMaybe}{$which}{pkts} += $pkts;
      if ($self->{NapUserMaybeFH}) {
	 syswrite($self->{NapUserMaybeFH}, $Cflow::raw, length $Cflow::raw)
      }
   }
   return 1
}

sub ASwanted {
   my $self = shift;
   my $ref = shift;
   my $aspair = "$src_as:$dst_as";
   $ref->{as}{$aspair}{flows}++;
   $ref->{as}{$aspair}{bytes} += $bytes;
   $ref->{as}{$aspair}{pkts} += $pkts
}

sub wanted_app {
   my $self = shift @_;

   # Trying to identifying applications is tricky business...
   # These tests should be mutually excusive (I think) and they should
   # be ordered from most to least confidence based on the method used
   # to identify the application:
   $self->ftpPASVWanted($self->{CampusIO}{which}, $self->{total}) or
   $self->RealWanted($self->{CampusIO}{which}, $self->{total}) or
   $self->QuakeWanted($self->{CampusIO}{which}, $self->{total}) or
   (@CampusIO::napster_files &&
    $self->NapsterWanted($self->{CampusIO}{which}, $self->{total}))
}

sub wanted_service {
   my $self = shift @_;

   # keep totals by service
   if (defined $CampusIO::service{$protocol}) {
      my $ref = $self->{total};
      if (defined $CampusIO::service{$protocol}{$srcport}) {
	  $ref->{service}{$protocol}{src}{$srcport}{$self->{CampusIO}{which}}{flows}++;
	  $ref->{service}{$protocol}{src}{$srcport}{$self->{CampusIO}{which}}{bytes} += $bytes;
	  $ref->{service}{$protocol}{src}{$srcport}{$self->{CampusIO}{which}}{pkts} += $pkts;
	  return 1
      }
      if (defined $CampusIO::service{$protocol}{$dstport}) {
	 $ref->{service}{$protocol}{dst}{$dstport}{$self->{CampusIO}{which}}{flows}++;
	 $ref->{service}{$protocol}{dst}{$dstport}{$self->{CampusIO}{which}}{bytes} += $bytes;
	 $ref->{service}{$protocol}{dst}{$dstport}{$self->{CampusIO}{which}}{pkts} += $pkts;
	 return 1
      }
   }
   return 0
}
sub wanted {
   my $self = shift;
   my $ref;

   $self->{CampusIO}{which} = ''; # unknown if this is an in or out-bound flow

   # check for multicast:
   if ($CampusIO::MCAST_NET == ($CampusIO::MCAST_MASK & $dstaddr)) {
      my $srcnet = $CampusIO::ptrie->match_integer($srcaddr);
      # FIXME?  What if multicast flow is intracampus (neither 'in' nor 'out')?:
      $self->{CampusIO}{which} = (ref($srcnet))? 'out' : 'in';
      # keep multicast grand totals...
      $self->{multicast}{total}{$self->{CampusIO}{which}}{flows}++;
      $self->{multicast}{total}{$self->{CampusIO}{which}}{bytes} += $bytes;
      $self->{multicast}{total}{$self->{CampusIO}{which}}{pkts} += $pkts;

      # return zero since we don't really know whether or not this multicast flows really represents Campus I/O or not
      return 0
   }

   if (0 == $nexthop) { # skip non-routable unicast traffic
      return 0
   }

   # FIXME - keep stats for flows where 0 == $output_if && 0 != $nexthop, then:
   return(0) if (0 == $output_if); # skip black-holed unicast traffic
   
   # maintain AS matrices
   # this isn't strictly "Campus I/O" - here we accumulate totals for
   # everything the router saw (that had a nexthop)...
   $self->ASwanted($self->{total});

   if ($CampusIO::hops_ptrie->match_integer($nexthop) ||
       grep($output_if == $_, @CampusIO::output_if)) {
      # this looks like its an outbound flow...
      $self->{CampusIO}{which} = 'out';

      # keep track of which interface it went out on...
      # Later, this will help us to determine if a other flows represents
      # inbound traffic.  (I.e. if a flow's input_if is one of the output_ifs
      # that was used to get to one of the @hops, then it is likely to be an 
      # inbound flow.  No?)
      $exporterip && $self->{if}->{$exporterip}{$self->{CampusIO}{which}}{$output_if}++;

      # keep outbound grand totals...
      $self->{total}->{$self->{CampusIO}{which}}{flows}++;
      $self->{total}->{$self->{CampusIO}{which}}{bytes} += $bytes;
      $self->{total}->{$self->{CampusIO}{which}}{pkts} += $pkts;

      # keep totals by proto
      if (defined $CampusIO::proto{$protocol}) {
	 $ref = $self->{total};
	 $ref->{proto}{$protocol}{$self->{CampusIO}{which}}{flows}++;
	 $ref->{proto}{$protocol}{$self->{CampusIO}{which}}{bytes} += $bytes;
	 $ref->{proto}{$protocol}{$self->{CampusIO}{which}}{pkts} += $pkts
      }

      my $identified = $self->wanted_app;
      if ($self->wanted_service) {
         $identified = 1
      }
      if (!$identified && $self->{otherFH}) {
	 syswrite($self->{otherFH}, $Cflow::raw, length $Cflow::raw)
      }

      my $cargo = $CampusIO::ptrie->match_integer($srcaddr);
      if (!ref($cargo)) { # keep outbound totals for unknown nets
         $self->{unknown}->{$self->{CampusIO}{which}}{flows}++;
         $self->{unknown}->{$self->{CampusIO}{which}}{bytes} += $bytes;
         $self->{unknown}->{$self->{CampusIO}{which}}{pkts} += $pkts;
         return(0)
      }

      # keep outbound totals for subnet...
      $cargo->{$self->{CampusIO}{which}}{bytes} += $bytes;
      $cargo->{$self->{CampusIO}{which}}{pkts} += $pkts;
      $cargo->{$self->{CampusIO}{which}}{flows}++;
      return 1
   } else { # Hmm, this *might* be an inbound flow...
      # be sure its nexthop is not another local router, lest we count the
      # traffic twice.
      if ($CampusIO::localhops_ptrie->match_integer($nexthop)) {
	 warn "Skipping \"inbound\" candidate flow from ${srcip} destined for ${dstip} via \"local\" nexthop ${nexthopip}.\n" if $CampusIO::Verbose;
	 return(0)
      }
      my $srcnet = $CampusIO::ptrie->match_integer($srcaddr);
      if (!ref($srcnet)) { # looks like it's a flow from an outside network...
         $self->{CampusIO}{which} = 'in';
         # keep inbound grand totals...
         $self->{total}->{$self->{CampusIO}{which}}{flows}++;
         $self->{total}->{$self->{CampusIO}{which}}{bytes} += $bytes;
         $self->{total}->{$self->{CampusIO}{which}}{pkts} += $pkts;

      # keep totals by proto
      if (defined $CampusIO::proto{$protocol}) {
	 $ref = $self->{total};
	 $ref->{proto}{$protocol}{$self->{CampusIO}{which}}{flows}++;
	 $ref->{proto}{$protocol}{$self->{CampusIO}{which}}{bytes} += $bytes;
	 $ref->{proto}{$protocol}{$self->{CampusIO}{which}}{pkts} += $pkts
      }

      my $identified = $self->wanted_app;
      if ($self->wanted_service) {
         $identified = 1
      }
      if (!$identified && $self->{otherFH}) {
	 syswrite($self->{otherFH}, $Cflow::raw, length $Cflow::raw)
      }

	 # keep a count of how many times we see this input interface
         $exporterip && $self->{if}->{$exporterip}{$self->{CampusIO}{which}}{$input_if}++;

         my $cargo = $CampusIO::ptrie->match_integer($dstaddr);
         if (!ref($cargo)) { # keep inbound totals for unknown nets
            $self->{unknown}->{$self->{CampusIO}{which}}{flows}++;
            $self->{unknown}->{$self->{CampusIO}{which}}{bytes} += $bytes;
            $self->{unknown}->{$self->{CampusIO}{which}}{pkts} += $pkts;
         } else {
            # keep inbound totals for subnet...
            $cargo->{$self->{CampusIO}{which}}{bytes} += $bytes;
            $cargo->{$self->{CampusIO}{which}}{pkts} += $pkts;
            $cargo->{$self->{CampusIO}{which}}{flows}++;
         }
	 return 1
      } else {
	 # even though this doesn't look like an inbound flow, check which
	 # interface it came in on and warn if it looks like it should have
	 # been selected as an inbound flow...
	 if ($CampusIO::Verbose && $exporterip && $self->{if}->{$exporterip}{out}{$input_if}) {
	    warn "Skipping flow from ${srcip} destined for ${dstip} via nexthop ${nexthopip} because it looks like an inbound flow, but is sourced from a local network/subnet.  It came in via interface $input_if (on router $exporterip), which is one of this router's known interfaces to the outside world: ", join(", ", keys(%{$self->{if}->{$exporterip}{out}})), ".\n"
	 }
	 return 0
      }
      # NOTREACHED
      die
   }
   # NOTREACHED
   die
}

=item perfile

=cut

sub perfile {
   my $self = shift;
   my $file = shift;

   $self->SUPER::perfile($file);

   if ('' eq ${CampusIO::outputdir}) { # write to the same directory
      $self->{outputdir} = dirname($file)
   } elsif (${CampusIO::outputdir} =~ m|^/|) { # write to the absolute directory
      $self->{outputdir} = ${CampusIO::outputdir}
   } else { # write to the relative directory
      $self->{outputdir} = dirname($file) . '/' . ${CampusIO::outputdir}
   }

   # Purge "old" entries from the RealServer hash:
   $self->RealServerPurge;

   # Purge "old" entries from the FTPSession hash:
   $self->FTPSessionPurge;

   if (@CampusIO::NapsterPorts) {
      # before processing the file, clear the NapsterCache of "old" entries:
      $self->NapsterCachePurge; # relies on filetime being set...

      my $dirname = dirname $file;
      my $basename = basename $file;
      for my $bin (qw(ftpPASV NapUser NapUserMaybe other)) {
         undef $self->{"${bin}FH"}; # close the previous raw flow file
         if (-d "$dirname/saved/${bin}") {
	    my $filename = "$dirname/saved/${bin}/$basename";
	    if (!($self->{"${bin}FH"} = new IO::File ">${filename}")) {
               warn(strftime("%Y/%m/%d %H:%M:%S", localtime),
		    " open \"$filename\" failed: $!\n")
            }
         }
      }
   }
}

sub NapsterCachePurge {
   my $self = shift;
   my $whence = $self->{filetime} - $CampusIO::NapsterSeconds;

   warn(strftime("%Y/%m/%d %H:%M:%S", localtime),
	" %CampusIO::NapServer -> ",
	scalar(%CampusIO::NapServer),
	" %CampusIO::NapUser -> ",
	scalar(%CampusIO::NapUser),
	"\n") if $CampusIO::verbose;

   while (my($k, $v) = each(%CampusIO::NapServer)) {
      if ($v < $whence) {
         warn("Purging NapServer ", inet_ntoa(pack("N", $k)), " ",
	      scalar(strftime("%Y/%m/%d %H:%M:%S", localtime($v))), " < ",
	      scalar(strftime("%Y/%m/%d %H:%M:%S", localtime($whence))), "\n")
		 if $CampusIO::Verbose;
         delete($CampusIO::NapServer{$k})
      }
   }
   while (my($k, $v) = each(%CampusIO::NapUser)) {
      if ($v < $whence) {
         warn("Purging NapUser ", inet_ntoa(pack("N", $k)), " ",
	      scalar(strftime("%Y/%m/%d %H:%M:%S", localtime($v))), " < ",
	      scalar(strftime("%Y/%m/%d %H:%M:%S", localtime($whence))), "\n")
		 if $CampusIO::Verbose;
         delete($CampusIO::NapUser{$k})
      }
   }

   warn(strftime("%Y/%m/%d %H:%M:%S", localtime),
	" %CampusIO::NapServer -> ",
	scalar(%CampusIO::NapServer),
	" %CampusIO::NapUser -> ",
	scalar(%CampusIO::NapUser),
	"\n") if $CampusIO::verbose;
}

sub FTPSessionPurge {
   my $self = shift;

   # For the timers below, experimentation with ~ an hour of flows on
   # 2000/07/18 ~23:00:00 CDT showed that I counted about 90% the total
   # number of PASV mode ftp traffic in bytes when the timeouts were
   # 5 and 15 minutes rather than 15 and 30, respectively.

   # purge sessions for which we've seen FIN after only 15 minutes:
   my $whence_fin = $self->{filetime} - 15*60;
   # purge sessions for which we've seen ACK (but no FIN) after 30 minutes:
   my $whence_ack = $self->{filetime} - 30*60;

   warn(strftime("%Y/%m/%d %H:%M:%S", localtime),
	" %CampusIO::FTPSession -> ",
	scalar(%CampusIO::FTPSession), "\n") if $CampusIO::verbose;

   while (my($k, $v) = each(%CampusIO::FTPSession)) {
      if ((-1 != $v->[1] && $v->[1] < $whence_fin) ||
	  $v->[0] < $whence_ack) {
         delete($CampusIO::FTPSession{$k})
      }
   }

   warn(strftime("%Y/%m/%d %H:%M:%S", localtime),
	" %CampusIO::FTPSession -> ",
	scalar(%CampusIO::FTPSession), "\n") if $CampusIO::verbose;
}

sub RealServerPurge {
   my $self = shift;
   # purge Real servers after 60 minutes of inactivity:
   my $whence = $self->{filetime} - 60*60;

   while (my($k, $v) = each(%CampusIO::RealServer)) {
      delete($CampusIO::RealServer{$k}) if ($v < $whence)
   }
}

=item report

=cut

sub createRRD {
   my $self = shift;
   die unless ref($self);
   my $file = shift;
   if (@_) {
      @CampusIO::prefix = @_
   } else {
      @CampusIO::prefix = ('')
   }

   my $time_t = $self->{filetime};

   my $startwhen = $time_t - 300;
   my(@DS, $prefix, $thingy);
   foreach $thingy ('bytes', 'pkts', 'flows') {
      foreach $prefix (@CampusIO::prefix) {
         push(@DS, "DS:${prefix}${thingy}:ABSOLUTE:400:U:U")
      }
   }
   RRDs::create($file,
      '--start', $startwhen,
      '--step', 300,
      @DS,
      qw(
         RRA:AVERAGE:0:1:600
         RRA:AVERAGE:0:6:600
         RRA:AVERAGE:0:24:600
         RRA:AVERAGE:0:288:732
         RRA:MAX:0:24:600
         RRA:MAX:0:288:732
	 )
      );
   my $err=RRDs::error;
   warn "ERROR creating $file: $err\n" if $err;
}

sub updateRRD {
   my $self = shift;
   die unless ref($self);
   my $file = shift;
   my @values = @_;

   RRDs::update($file, $self->{filetime} . ':' . join(':', @values));
   my $err=RRDs::error;
   if ($err) {
      warn "ERROR updating $file: $err\n"
   }
}

sub MulticastReportRRD {
   my $self = shift;
   my $file = $self->{outputdir} . "/MCAST.rrd";
   $self->createRRD($file, 'in_', 'out_') unless -f $file;

   my @values = ();
   my $thingy;
   foreach $thingy ('bytes', 'pkts', 'flows') {
      push(@values,
	   0+$self->{multicast}{total}{in}{$thingy},
	   0+$self->{multicast}{total}{out}{$thingy}
	  );
   }
   $self->updateRRD($file, @values)
}

sub RealReportRRD {
   my $self = shift;
   my $file = $self->{outputdir} . "/RealAudio.rrd";
   $self->createRRD($file, 'in_', 'out_') unless -f $file;

   my @values = ();
   my $thingy;
   foreach $thingy ('bytes', 'pkts', 'flows') {
      push(@values,
	   0+$self->{total}->{app}{RealAudio}{in}{$thingy},
	   0+$self->{total}->{app}{RealAudio}{out}{$thingy}
	  );
   }
   $self->updateRRD($file, @values);

   warn(strftime("%Y/%m/%d %H:%M:%S", localtime), " scalar(%CampusIO::RealServer) -> ", scalar(%CampusIO::RealServer), "\n") if $CampusIO::verbose;
}

sub AppReportRRD {
   my $self = shift;
   my $app = shift;
   my $file = $self->{outputdir} . "/${app}.rrd";
   $self->createRRD($file, 'in_', 'out_') unless -f $file;

   my @values = ();
   my $thingy;
   foreach $thingy ('bytes', 'pkts', 'flows') {
      push(@values,
	   0+$self->{total}->{app}{$app}{in}{$thingy},
	   0+$self->{total}->{app}{$app}{out}{$thingy}
	  );
   }
   $self->updateRRD($file, @values);
}

sub ASreportRRD {
   my $self = shift;
   my $aspair;

   foreach $aspair (@CampusIO::as2as) {
      my $file = $self->{outputdir} . "/$aspair.rrd";
      $self->createRRD($file, '') unless -f $file;

      my @values = ();
      my $thingy;
      foreach $thingy ('bytes', 'pkts', 'flows') {
         push(@values, 0+$self->{total}->{as}{$aspair}{$thingy});
      }
      $self->updateRRD($file, @values);
   }
}

sub reportRRD {
   my $self = shift;
   $self->ASreportRRD;
   $self->AppReportRRD('ftpPASV_src');
   $self->AppReportRRD('ftpPASV_dst');
   $self->RealReportRRD;
   $self->AppReportRRD('Quake');
   $self->AppReportRRD('NapUser') if @CampusIO::napster_files;
   $self->AppReportRRD('NapUserMaybe') if @CampusIO::napster_files;
   $self->MulticastReportRRD;
   if (@CampusIO::subnets) {
      foreach (@CampusIO::subnets) {
         my $subnet = $_->{SUBNET};
         $subnet =~ s:/:_:; # "/" not allowed in file names...
         my $file = $self->{outputdir} . "/${subnet}.rrd";
         $self->createRRD($file, 'in_', 'out_') unless -f $file;
         my @values = ();
	 my $thingy;
	 foreach $thingy ('bytes', 'pkts', 'flows') {
	    push(@values, 0+$_->{in}{$thingy}, 0+$_->{out}{$thingy});
	 }
         $self->updateRRD($file, @values);
      }
   }
   # { do totals and unknown
   my $grand; # a grand total
   foreach $grand ('total', 'unknown') {
      my $file = $self->{outputdir} . "/${grand}.rrd";
      $self->createRRD($file, 'in_', 'out_') unless -f $file;

      my @values = ();
      my $thingy;
      foreach $thingy ('bytes', 'pkts', 'flows') {
         push(@values,
	      0+$self->{$grand}->{in}{$thingy},
	      0+$self->{$grand}->{out}{$thingy}
	     );
      }
      $self->updateRRD($file, @values);
   }

   # }{ do protocols

   { # for scope only
   my $proto;
   foreach $proto (keys %CampusIO::proto) {
      my $name = getprotobynumber($proto);
      if ('' eq $name) {
	 $name = $proto
      }
      my $file = $self->{outputdir} . "/${name}.rrd";
      $self->createRRD($file, 'in_', 'out_') unless -f $file;

      my @values = ();
      my $thingy;
      foreach $thingy ('bytes', 'pkts', 'flows') {
         push(@values,
	      0+$self->{total}->{proto}{$proto}{in}{$thingy},
	      0+$self->{total}->{proto}{$proto}{out}{$thingy}
	     );
      }
      $self->updateRRD($file, @values);
   }
   }

   # }{ do services
   my($proto, $port, $direction);
   foreach $proto (keys %{$self->{total}->{service}}) {
      my $protoname = getprotobynumber($proto);
      die if ('' eq $protoname);
      foreach $direction ('src', 'dst') {
	 if ('tcp' eq $protoname ||
	     'udp' eq $protoname) {
	 foreach $port (keys %{$self->{total}->{service}{$proto}{$direction}}) {
	    my $service = getservbyport($port, $protoname);
	    if ('' eq $service) {
	       $service = $port
	    }

            my $file = $self->{outputdir} . "/"; # to be continued...
	    if ('tcp' eq $protoname) {
	       # For now keep the old file names for backward compatibility
	       # with RRD files that were created back when there were only
	       # TCPServices (no UDPServices).  FIXME
	       $file .= "${service}_${direction}.rrd"
	    } else {
	       $file .= "${protoname}_${service}_${direction}.rrd"
	    }

            $self->createRRD($file, 'in_', 'out_') unless -f $file;

            my @values = ();
            my $thingy;
            foreach $thingy ('bytes', 'pkts', 'flows') {
             push(@values,
	    0+$self->{total}->{service}{$proto}{$direction}{$port}{in}{$thingy},
	    0+$self->{total}->{service}{$proto}{$direction}{$port}{out}{$thingy}
	     );
            }
            $self->updateRRD($file, @values);
         }
	 } elsif ('icmp' eq $protoname) {
	    # FIXME - add the reporting of ICMP by type/code
	 }
      }
   }
   # }
}

sub report {
   my $self = shift;
   $self->reportRRD;
   return unless $CampusIO::rateup;
   return unless @CampusIO::subnets;
   foreach $CampusIO::thingy (${CampusIO::thing}? (${CampusIO::thing}) : ('flows', 'pkts', 'bytes')) {
      foreach (@CampusIO::subnets) {
	 my $subnet = $_->{SUBNET};
	 $subnet =~ s:/:_:;
	 my $invalue = $_->{in}{$CampusIO::thingy};
	 my $outvalue = $_->{out}{$CampusIO::thingy};
         $self->dorateup($invalue, $outvalue, $subnet, $CampusIO::thingy)
      }
      # { do totals
      my $invalue = $self->{total}->{in}{$CampusIO::thingy};
      my $outvalue = $self->{total}->{out}{$CampusIO::thingy};
      $self->dorateup($invalue, $outvalue, 'total', $CampusIO::thingy);
      # }{ do unknown
      my $invalue = $self->{unknown}->{in}{$CampusIO::thingy};
      my $outvalue = $self->{unknown}->{out}{$CampusIO::thingy};
      $self->dorateup($invalue, $outvalue, 'unknown', $CampusIO::thingy);
      # }{ do services
	 my($proto, $port, $direction);
	 foreach $proto (keys %{$self->{total}->{service}}) {
	    foreach $direction ('src', 'dst') {
	       foreach $port (keys %{$self->{total}->{service}{$proto}{$direction}}) {
                  my $invalue = $self->{total}->{service}{$proto}{$direction}{$port}{in}{$CampusIO::thingy};
                  my $outvalue = $self->{total}->{service}{$proto}{$direction}{$port}{out}{$CampusIO::thingy};
		  # FIXME ('tcp' shouldn't be hardcoded here):
		  my $service = getservbyport($port, 'tcp');
	          if ('' eq $service) {
	             $service = $port
	          }
		  $self->dorateup($invalue, $outvalue, "total_${service}_${direction}", $CampusIO::thingy)
	       }
	    }
	 }
      # }
   }
}

sub dorateup {
   my $self = shift;
   die unless ref($self);
   # use rateup's 'a'bsolute option here so that it doesn't mess with
   # our values (by subtracting the previous value).  (We would use
   # rateup's 'g'auge option here if we didn't want it to mess with our
   # values by subtracting the previous value or dividing by the period.)
   my $type = 'a';
   my $invalue = shift;
   my $outvalue = shift;
   my $title = shift;
   my $thingy = shift;
   my $command = sprintf("${CampusIO::rateup} $self->{outputdir} ${title}_${thingy} $self->{filetime} ${type} %.0f %.0f %u", ${invalue}, ${outvalue}, &UINT_MAX);
   if (1) { # produce a graph
      $command .= " l '[${thingy} per second]' p $self->{outputdir}/${title}_${thingy}.gif 100 400 100 1 1 0 300 0";
   }
   my $status = int(system($command)/256);
   if (0 != $status) {
      warn "\"${command}\" finished with exit status: ${status}\n"
   }
}

sub DESTROY {
   # zero the subnet counters (since they are class-level rather than
   # instance-level data objects)
   my $subnet;
   foreach $subnet (@CampusIO::subnets) {
      delete $subnet->{in};
      delete $subnet->{out}
   }
}

=head1 BUGS

=head1 AUTHOR

Dave Plonka <plonka@doit.wisc.edu>

Copyright (C) 1998-2000  Dave Plonka.
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

=head1 VERSION

The version number is the module file RCS revision number (B<$Revision: 1.45 $>)
with the minor number printed right justified with leading zeroes to 3
decimal places.  For instance, RCS revision 1.1 would yield a package version
number of 1.001.

This is so that revision 1.10 (which is version 1.010), for example, will
test greater than revision 1.2 (which is version 1.002) when you want to
B<require> a minimum version of this module.

=cut

1
