#  SubNetIO.pm - a FlowScan-derived class for reporting on traffic I/O
#  Copyright (C) 1999-2000  Dave Plonka
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

# $Id: SubNetIO.pm,v 1.14 2000/10/18 21:03:02 dplonka Exp $
# Dave Plonka <plonka@doit.wisc.edu>

use strict;

package SubNetIO;

require 5;
require Exporter;

use CampusIO 1.016;

@SubNetIO::ISA=qw(CampusIO Exporter);
# convert the RCS revision to a reasonable Exporter VERSION:
'$Revision: 1.14 $' =~ m/(\d+)\.(\d+)/ && (( $SubNetIO::VERSION ) = sprintf("%d.%03d", $1, $2));

=head1 NAME

SubNetIO -

=head1 SYNOPSIS

   $ flowscan SubNetIO

=head1 DESCRIPTION

The following methods and subroutines are defined:

=over 4

=cut

use Cflow qw(:flowvars 1.015); # for use in wanted sub
use RRDs;
use Boulder::Stream;
use IO::File;
use File::Basename;
use POSIX; # for mktime
use ConfigReader::DirectiveStyle;
use Net::Patricia 1.008;

my $c = new ConfigReader::DirectiveStyle;
$c->required('OutputDir');
$c->directive('SubnetFiles');
$c->directive('Verbose');
$c->load("${FindBin::Bin}/SubNetIO.cf");

if (1 <= $c->value('Verbose')) {
   $SubNetIO::verbose = 1
}
if (2 <= $c->value('Verbose')) {
   $SubNetIO::Verbose = 1
}

# outputdir can be absolute or relative (to the flow file's directory):
$SubNetIO::outputdir = $c->value('OutputDir');

$SubNetIO::thingy; # this is a global set by report subroutine and used by the sort subs
$SubNetIO::thing = 'bytes';

@SubNetIO::subnet_files = split(m/\s*,\s*/, $c->value('SubnetFiles'));

# { initialize the Patricia Trie
$SubNetIO::patricia = new Net::Patricia;
die unless ref($SubNetIO::patricia);

@SubNetIO::subnets_files = <@SubNetIO::subnet_files>;
my($subnets_file, $stream, $cargo);
foreach $subnets_file (@SubNetIO::subnets_files) {
   print(STDERR "Loading \"$subnets_file\" ...\n") if -t;
   my $fh = new IO::File "<$subnets_file";
   $fh || die "open \"$subnets_file\", \"r\": $!\n";
   $stream = new Boulder::Stream $fh;
   while ($cargo = $stream->read_record) {
      my $subnet = $cargo->get('SUBNET');
      die unless $subnet;
      my $collision;
      if ($collision = $SubNetIO::patricia->match_string($subnet)) {
         warn "$subnet patricia->add skipped - collided with $collision->{SUBNET}\n";
	 next;
      }
      my $hr = { SUBNET => $subnet };
      if (!$SubNetIO::patricia->add_string($subnet, $hr)) {
         warn "$subnet patricia->add failed!\n";
      }
   }
   undef $fh
}

# }

=item new

The B<new> method constructs and returns a B<SubNetIO> object.

=cut

sub new {
   my $self = {};
   my $class = shift;
   CampusIO::_init($self);
   return bless _init($self), $class
}

sub _init {
   my $self = shift;
   return $self
}

=item wanted

=cut

sub wanted {
   my $self = shift;
   my $ref;
   my $which; # 'in' or 'out'

   my $rv = $self->SUPER::wanted;
   return($rv) unless $rv;
   
   if ('out' eq $self->{CampusIO}{which}) { # looks like it's an outbound flow
      $which = 'out';
      $cargo = $SubNetIO::patricia->match_integer($srcaddr);
   } else { # Hmm, this is an inbound flow...
      die unless 'in' eq $self->{CampusIO}{which};
      $which = 'in';
      $cargo = $SubNetIO::patricia->match_integer($dstaddr);
   }

   return($rv) unless ref($cargo);

   $cargo->{$which}{bytes} += $bytes;
   $cargo->{$which}{pkts} += $pkts;
   $cargo->{$which}{flows}++;

   return($rv)
}

=item perfile

=cut

sub perfile {
   my $self = shift;
   my $file = shift;
   $self->SUPER::perfile($file);
   if ('' eq ${SubNetIO::outputdir}) { # write to the same directory
      $self->{outputdir} = dirname($file)
   } elsif (${SubNetIO::outputdir} =~ m|^/|) { # write to the absolute directory
      $self->{outputdir} = ${SubNetIO::outputdir}
   } else { # write to the relative directory
      $self->{outputdir} = dirname($file) . '/' . ${SubNetIO::outputdir}
   }
}

=item report

=cut

sub report_node {
   my $self = shift(@_);
   die unless ref($self);
   my $hr = shift(@_);
   die unless ref($hr);

   my $subnet = $hr->{SUBNET};
   die unless $subnet;
   $subnet =~ s:/:_:; # "/" not allowed in file names...
   my $file = $self->{outputdir} . "/${subnet}.rrd";
   $self->createRRD($file, 'in_', 'out_') unless -f $file;
   my @values = ();
   foreach my $thingy ('bytes', 'pkts', 'flows') {
      push(@values, 0+$hr->{in}{$thingy}, 0+$hr->{out}{$thingy});
   }
   $self->updateRRD($file, @values)
}

sub report {
   my $self = shift;
   $self->SUPER::report;
   $SubNetIO::patricia->climb(sub { $self->report_node(@_) })
}

sub clear_node {
   my $hr = shift(@_);
   die unless ref($hr);
   foreach my $which ('in', 'out') {
      foreach my $thingy ('bytes', 'pkts', 'flows') {
         $hr->{$which}{$thingy} = 0
      }
   }
}

sub DESTROY {
   my $self = shift;
   # zero the subnet counters (since they are class-level rather than
   # instance-level data objects)
   $SubNetIO::patricia->climb(\&clear_node);
   $self->SUPER::DESTROY
}

=head1 BUGS

=head1 AUTHOR

Dave Plonka <plonka@doit.wisc.edu>

Copyright (C) 1999-2000  Dave Plonka.
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

=head1 VERSION

The version number is the module file RCS revision number (B<$Revision: 1.14 $>)
with the minor number printed right justified with leading zeroes to 3
decimal places.  For instance, RCS revision 1.1 would yield a package version
number of 1.001.

This is so that revision 1.10 (which is version 1.010), for example, will
test greater than revision 1.2 (which is version 1.002) when you want to
B<require> a minimum version of this module.

=cut

1
